#ifndef WIN32
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#endif
#include <msp/strings/format.h>
#include <msp/core/systemerror.h>
#include "file.h"
#include "handle_private.h"

using namespace std;

namespace Msp {
namespace IO {

File::File(const string &fn, Mode m, CreateMode cm)
{
	if(!(m&M_RDWR))
		throw invalid_argument("File::File mode");
	if(cm&~(C_CREATE|C_TRUNCATE))
		throw invalid_argument("File::File create");

	mode = m;

#ifdef WIN32
	int flags = 0;
	int create_flags = OPEN_EXISTING;

	if(mode&M_READ)
		flags |= GENERIC_READ;
	else if(mode&M_WRITE)
	{
		flags |= GENERIC_WRITE;

		switch(static_cast<int>(cm))
		{
		case C_NONE:     create_flags = OPEN_EXISTING; break;
		case C_CREATE:   create_flags = OPEN_ALWAYS; break;
		case C_TRUNCATE: create_flags = TRUNCATE_EXISTING; break;
		case C_CREATE+C_TRUNCATE: create_flags = CREATE_ALWAYS; break;
		}
	}

	*handle = CreateFile(fn.c_str(), flags, 0, 0, create_flags, FILE_ATTRIBUTE_NORMAL, 0);
	if(!handle)
	{
		int err = GetLastError();
		if(err==ERROR_FILE_NOT_FOUND)
			throw file_not_found(fn);
		else
			throw system_error(format("CreateFile(%s)", fn), err);
	}
#else
	int flags = 0;
	switch(mode&M_RDWR)
	{
	case M_READ:  flags |= O_RDONLY; break;
	case M_WRITE: flags |= O_WRONLY; break;
	case M_RDWR:  flags |= O_RDWR; break;
	default:;
	}

	if(mode&M_WRITE)
	{
		if(cm&C_CREATE)
			flags |= O_CREAT;
		if(cm&C_TRUNCATE)
			flags |= O_TRUNC;
	}
	if(mode&M_APPEND)
		flags |= O_APPEND;
	if(mode&M_NONBLOCK)
		flags |= O_NONBLOCK;

	*handle = ::open(fn.c_str(), flags, 0666);
	if(!handle)
	{
		int err = errno;
		if(err==ENOENT)
			throw file_not_found(fn);
		else
			throw system_error(format("open(%s)", fn), err);
	}
#endif
}

File::~File()
{
	signal_flush_required.emit();
	sys_close(handle);
}

void File::set_block(bool b)
{
	mode = (mode&~M_NONBLOCK);
	if(b)
		mode = (mode|M_NONBLOCK);
#ifndef WIN32
	int flags = fcntl(*handle, F_GETFD);
	fcntl(*handle, F_SETFL, (flags&O_NONBLOCK)|(b?0:O_NONBLOCK));
#endif
}

unsigned File::do_write(const char *buf, unsigned size)
{
	check_access(M_WRITE);

	if(size==0)
		return 0;

#ifdef WIN32
	if(mode&M_APPEND)
		seek(0, S_END);
#endif

	return sys_write(handle, buf, size);
}

unsigned File::do_read(char *buf, unsigned size)
{
	check_access(M_READ);

	if(size==0)
		return 0;

	unsigned ret = sys_read(handle, buf, size);
	if(ret==0)
		set_eof();

	return ret;
}

void File::sync()
{
#ifndef WIN32
	signal_flush_required.emit();

	fsync(*handle);
#endif
}

SeekOffset File::seek(SeekOffset off, SeekType type)
{
	signal_flush_required.emit();
	off = sys_seek(handle, off, type);
	eof_flag = false;

	return off;
}

SeekOffset File::tell() const
{
	return sys_seek(const_cast<Handle &>(handle), 0, S_CUR);
}

const Handle &File::get_handle(Mode m)
{
	check_access(m);
	return handle;
}

} // namespace IO
} // namespace Msp
